/*
 *  linux/arch/i386/kernel/time.c
 *
 *  Copyright (C) 1991, 1992, 1995  Linus Torvalds
 *
 * This file contains the PC-specific time handling details:
 * reading the RTC at bootup, etc..
 * 1994-07-02    Alan Modra
 *	fixed set_rtc_mmss, fixed time.year for >= 2000, new mktime
 * 1995-03-26    Markus Kuhn
 *      fixed 500 ms bug at call to set_rtc_mmss, fixed DS12887
 *      precision CMOS clock update
 * 1996-05-03    Ingo Molnar
 *      fixed time warps in do_[slow|fast]_gettimeoffset()
 * 1997-09-10	Updated NTP code according to technical memorandum Jan '96
 *		"A Kernel Model for Precision Timekeeping" by Dave Mills
 */
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/param.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/interrupt.h>
#include <linux/time.h>
#include <linux/delay.h>

#ifdef CONFIG_FAST_TIMER
#include "linux/fast_timer.h"
#endif

#include <asm/segment.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/delay.h>

#include <linux/mc146818rtc.h>
#include <linux/timex.h>
#include <linux/config.h>

extern int setup_x86_irq(int, struct irqaction *);

#ifdef CONFIG_FAST_TIMER
/* With the fast timer, we store the number of "mini-interrupts"
 * (fast interrupts) left until the next HZ run.  With this
 * value we can also fabricate a big counter value for code
 * that wants it (the "slow" (8253-based) clock offset method).
 */
static int miniinterrupts;	/* a value between 0 and FAST_FACTOR */

/* This version incorporates some later fixes by Ingo Molnar
 * from the 2.1.XX code line and Ulrich Windl's "Completely Mad Time."
 * (See comments below.)  These seem to improve the clock behavior
 * considerably, especially when the "fast" (Pentium cycle counter-based)
 * clock offset method is used.  However, there can still be some remaining
 * irregularities.   These problems exist independently of the "fast timer,"
 * though admittedly the latter can exacerbate them.
 */

/* We keep a local "lost tick" count since the fast timer routines run
 * before do_timer sets the global lost tick count.  Local lost ticks
 * aren't really "lost," just not updated yet.  It's also not really
 * a count, but just a boolean indication of whether we've updated the
 * global xtime yet.
 */
static local_lost_ticks, local_lost_minis;

#endif

#ifndef	CONFIG_APM	/* cycle counter may be unreliable */
/* Cycle counter value at the previous timer interrupt.. */
static struct {
	unsigned long low;
	unsigned long high;
} init_timer_cc, last_timer_cc;

extern volatile unsigned long lost_ticks;


/* change this if you have some constant time drift */
#define USECS_PER_JIFFY (1000020/HZ)

/*
 * Cached "1/(clocks per usec)*2^32" value. 
 * It has to be recalculated once each jiffy.
 */
static unsigned long cached_quotient=0;

static unsigned long do_fast_gettimeoffset(void)
{
	register unsigned long eax asm("ax");
	register unsigned long edx asm("dx");
	unsigned long tmp, quotient, low_timer;

	/* Last jiffy when do_fast_gettimeoffset() was called. */
	static unsigned long last_jiffies=0;

	tmp = jiffies;

	quotient = cached_quotient;
	low_timer = last_timer_cc.low;

	if (last_jiffies != tmp) {
		last_jiffies = tmp;

		/* Get last timer tick in absolute kernel time */
		eax = low_timer;
		edx = last_timer_cc.high;
		__asm__("subl "SYMBOL_NAME_STR(init_timer_cc)",%0\n\t"
			"sbbl "SYMBOL_NAME_STR(init_timer_cc)"+4,%1"
			:"=a" (eax), "=d" (edx)
			:"0" (eax), "1" (edx));

		/*
		 * Divide the 64-bit time with the 32-bit jiffy counter,
		 * getting the quotient in clocks.
		 *
		 * Giving quotient = "1/(average internal clocks per usec)*2^32"
		 * we do this '1/...' trick to get the 'mull' into the critical 
		 * path. 'mull' is much faster than divl (10 vs. 41 clocks)
		 */
		__asm__("divl %2"
			:"=a" (eax), "=d" (edx)
			:"r" (tmp),
			 "0" (eax), "1" (edx));

		edx = USECS_PER_JIFFY;
		tmp = eax;
		eax = 0;

		__asm__("divl %2"
			:"=a" (eax), "=d" (edx)
			:"r" (tmp),
			 "0" (eax), "1" (edx));
		cached_quotient = eax;
		quotient = eax;
	}

	/* Read the time counter */
	__asm__(".byte 0x0f,0x31"
		:"=a" (eax), "=d" (edx));


	/* .. relative to previous jiffy (32 bits is enough) */
	edx = 0;
	eax -= low_timer;

	/*
	 * Time offset = (USECS_PER_JIFFY * time_low) * quotient.
	 */

	__asm__("mull %2"
		:"=a" (eax), "=d" (edx)
		:"r" (quotient),
		 "0" (eax), "1" (edx));

#ifdef notdef
	/*
 	 * Due to possible jiffies inconsistencies, we need to check 
	 * the result so that we'll get a timer that is monotonic.
	 */
	if (edx >= USECS_PER_JIFFY)
		edx = USECS_PER_JIFFY-1;
#else
	/* CMT has its own way of making the timer monotonic.  See the
	 * comments below....
	 */
#endif

	return edx;
}
#endif

/* This function must be called with interrupts disabled 
 * It was inspired by Steve McCanne's microtime-i386 for BSD.  -- jrs
 * 
 * However, the pc-audio speaker driver changes the divisor so that
 * it gets interrupted rather more often - it loads 64 into the
 * counter rather than 11932! This has an adverse impact on
 * do_gettimeoffset() -- it stops working! What is also not
 * good is that the interval that our timer function gets called
 * is no longer 10.0002 ms, but 9.9767 ms. To get around this
 * would require using a different timing source. Maybe someone
 * could use the RTC - I know that this can interrupt at frequencies
 * ranging from 8192Hz to 2Hz. If I had the energy, I'd somehow fix
 * it so that at startup, the timer code in sched.c would select
 * using either the RTC or the 8253 timer. The decision would be
 * based on whether there was any other device around that needed
 * to trample on the 8253. I'd set up the RTC to interrupt at 1024 Hz,
 * and then do some jiggery to have a version of do_timer that 
 * advanced the clock by 1/1024 s. Every time that reached over 1/100
 * of a second, then do all the old code. If the time was kept correct
 * then do_gettimeoffset could just return 0 - there is no low order
 * divider that can be accessed.
 *
 * Ideally, you would be able to use the RTC for the speaker driver,
 * but it appears that the speaker driver really needs interrupt more
 * often than every 120 us or so.
 *
 * Anyway, this needs more thought....		pjsg (1993-08-28)
 * 
 * If you are really that interested, you should be reading
 * comp.protocols.time.ntp!
 */

#define TICK_SIZE tick

static unsigned long do_slow_gettimeoffset(void)
{
	int count;

	static int count_p = LATCH;    /* for the first call after boot */
	static unsigned long jiffies_p = 0;

	/*
	 * cache volatile jiffies temporarily; we have IRQs turned off. 
	 */
	unsigned long jiffies_t;

	/* timer count may underflow right here */
	outb_p(0x00, 0x43);	/* latch the count ASAP */

	count = inb_p(0x40);	/* read the latched count */

	/*
	 * We do this guaranteed double memory access instead of a _p 
	 * postfix in the previous port access. Wheee, hackady hack
	 */
 	jiffies_t = jiffies;

	count |= inb_p(0x40) << 8;

#ifdef CONFIG_FAST_TIMER
	/* Add in fast timer cycles */
	count += (miniinterrupts-local_lost_minis)*(LATCH/FAST_FACTOR);
#endif

	/*
	 * avoiding timer inconsistencies (they are rare, but they happen)...
	 * there are two kinds of problems that must be avoided here:
	 *  1. the timer counter underflows
	 *  2. hardware problem with the timer, not giving us continuous time,
	 *     the counter does small "jumps" upwards on some Pentium systems,
	 *     (see c't 95/10 page 335 for Neptun bug.)
	 */

/* you can safely undefine this if you dont have the Neptun chipset */

/*
 * @@ Note: (Mark Carson) The "Neptune" chipset is the Intel 430NX chipset.
 * It is used on some old 75-166MHz Pentium motherboards.  I don't know which
 * revs of the chipset have this bug.
 */

#define BUGGY_NEPTUN_TIMER

	if( jiffies_t == jiffies_p ) {
		if( count > count_p ) {
			/* the nutcase */

			outb_p(0x0A, 0x20);

			/* assumption about timer being IRQ1 */
			if( inb(0x20) & 0x01 ) {
				/*
				 * We cannot detect lost timer interrupts ... 
				 * well, thats why we call them lost, dont we? :)
				 * [hmm, on the Pentium and Alpha we can ... sort of]
				 */
#ifdef CONFIG_FAST_TIMER
				/* Note this correction can make the count
				 * go negative.  This means that we missed
				 * (or have not yet taken) the "major"
				 * clock interrupt.
				 */
				/* Todo: if we have a cycle counter, use it to
				 * determine/estimate the number of missed
				 * interrupts.  This is probably not too 
				 * important because (a) in practice, it
				 * doesn't look like more than one is ever
				 * missed, at least on a reasonably fast
				 * machine (b) if we have a cycle counter,
				 * we can use it directly (normally) instead
				 * of using this method.
				 */
				count -= LATCH/FAST_FACTOR;
#else
				count -= LATCH;
#endif
			} else {
#ifdef BUGGY_NEPTUN_TIMER
				/*
				 * for the Neptun bug we know that the 'latch'
				 * command doesnt latch the high and low value
				 * of the counter atomically. Thus we have to 
				 * substract 256 from the counter 
				 * ... funny, isnt it? :)
				 */

				count -= 256;
#else
				printk("do_slow_gettimeoffset(): hardware timer problem?\n");
#endif
			}
		}
	} else
		jiffies_p = jiffies_t;

	count_p = count;

	count = ((LATCH-1) - count) * TICK_SIZE;
	count = (count + LATCH/2) / LATCH;

	return count;
}

static unsigned long (*do_gettimeoffset)(void) = do_slow_gettimeoffset;


/*
 * This version of gettimeofday has near microsecond resolution.
 */
void do_gettimeofday(struct timeval *tv)
{
	unsigned long flags;

	save_flags(flags);
	cli();
	*tv = xtime;
	tv->tv_usec += do_gettimeoffset();

	/*
	 * xtime is atomically updated in timer_bh. lost_ticks is
	 * nonzero if the timer bottom half hasnt executed yet.
	 */
#ifdef CONFIG_FAST_TIMER
	if (lost_ticks || local_lost_ticks)
#else
	if (lost_ticks)
#endif
		tv->tv_usec += USECS_PER_JIFFY;

	restore_flags(flags);

	if (tv->tv_usec >= 1000000) {
		tv->tv_usec -= 1000000;
		tv->tv_sec++;
	}
}

void do_settimeofday(struct timeval *tv)
{
	cli();
	/* This is revolting. We need to set the xtime.tv_usec
	 * correctly. However, the value in this location is
	 * is value at the last tick.
	 * Discover what correction gettimeofday
	 * would have done, and then undo it!
	 */
	tv->tv_usec -= do_gettimeoffset();

	if (tv->tv_usec < 0) {
		tv->tv_usec += 1000000;
		tv->tv_sec--;
	}

	xtime = *tv;
	time_adjust = 0;		/* stop active adjtime() */
	time_status |= STA_UNSYNC;
	time_state = TIME_ERROR;	/* p. 24, (a) */
	time_maxerror = NTP_PHASE_LIMIT;
	time_esterror = NTP_PHASE_LIMIT;
	sti();
}


/*
 * In order to set the CMOS clock precisely, set_rtc_mmss has to be
 * called 500 ms after the second nowtime has started, because when
 * nowtime is written into the registers of the CMOS clock, it will
 * jump to the next second precisely 500 ms later. Check the Motorola
 * MC146818A or Dallas DS12887 data sheet for details.
 *
 * BUG: This routine does not handle hour overflow properly; it just
 *      sets the minutes. Usually you'll only notice that after reboot!
 */
static int set_rtc_mmss(unsigned long nowtime)
{
	int retval = 0;
	int real_seconds, real_minutes, cmos_minutes;
	unsigned char save_control, save_freq_select;

	save_control = CMOS_READ(RTC_CONTROL); /* tell the clock it's being set */
	CMOS_WRITE((save_control|RTC_SET), RTC_CONTROL);

	save_freq_select = CMOS_READ(RTC_FREQ_SELECT); /* stop and reset prescaler */
	CMOS_WRITE((save_freq_select|RTC_DIV_RESET2), RTC_FREQ_SELECT);

	cmos_minutes = CMOS_READ(RTC_MINUTES);
	if (!(save_control & RTC_DM_BINARY) || RTC_ALWAYS_BCD)
		BCD_TO_BIN(cmos_minutes);

	/*
	 * since we're only adjusting minutes and seconds,
	 * don't interfere with hour overflow. This avoids
	 * messing with unknown time zones but requires your
	 * RTC not to be off by more than 15 minutes
	 */
	real_seconds = nowtime % 60;
	real_minutes = nowtime / 60;
	if (((abs(real_minutes - cmos_minutes) + 15)/30) & 1)
		real_minutes += 30;		/* correct for half hour time zone */
	real_minutes %= 60;

	if (abs(real_minutes - cmos_minutes) < 30) {
		if (!(save_control & RTC_DM_BINARY) || RTC_ALWAYS_BCD) {
			BIN_TO_BCD(real_seconds);
			BIN_TO_BCD(real_minutes);
		}
		CMOS_WRITE(real_seconds,RTC_SECONDS);
		CMOS_WRITE(real_minutes,RTC_MINUTES);
	} else {
		printk(KERN_WARNING
		       "set_rtc_mmss: can't update from %d to %d\n",
		       cmos_minutes, real_minutes);
		retval = -1;
	}

	/* The following flags have to be released exactly in this order,
	 * otherwise the DS12887 (popular MC146818A clone with integrated
	 * battery and quartz) will not reset the oscillator and will not
	 * update precisely 500 ms later. You won't find this mentioned in
	 * the Dallas Semiconductor data sheets, but who believes data
	 * sheets anyway ...                           -- Markus Kuhn
	 */
	CMOS_WRITE(save_control, RTC_CONTROL);
	CMOS_WRITE(save_freq_select, RTC_FREQ_SELECT);

	return retval;
}

/* last time the cmos clock got updated */
static long last_rtc_update = 0;

#ifdef CONFIG_FAST_TIMER
extern void do_fast_timer(struct pt_regs *regs);
#endif

/*
 * timer_interrupt() needs to keep up the real-time clock,
 * as well as call the "do_timer()" routine every clocktick
 */
static inline void timer_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
#ifdef CONFIG_FAST_TIMER
	++local_lost_minis;	/* for the "slow" route */
	do_fast_timer(regs);
	if (miniinterrupts > 0) {
		--miniinterrupts;
		local_lost_minis = 0;
		return;
	}
	local_lost_minis = 0;
	local_lost_ticks = 0;
#endif
	do_timer(regs);

	/*
	 * If we have an externally synchronized Linux clock, then update
	 * CMOS clock accordingly every ~11 minutes. Set_rtc_mmss() has to be
	 * called as close as possible to 500 ms before the new second starts.
	 */
	if ((time_status & STA_UNSYNC) == 0 &&
	    xtime.tv_sec > last_rtc_update + 660 &&
	    xtime.tv_usec > 500000 - (tick >> 1) &&
	    xtime.tv_usec < 500000 + (tick >> 1))
	  if (set_rtc_mmss(xtime.tv_sec) == 0)
	    last_rtc_update = xtime.tv_sec;
	  else
	    last_rtc_update = xtime.tv_sec - 600; /* do it again in 60 s */
	/* As we return to user mode fire off the other CPU schedulers.. this is 
	   basically because we don't yet share IRQ's around. This message is
	   rigged to be safe on the 386 - basically it's a hack, so don't look
	   closely for now.. */
	/*smp_message_pass(MSG_ALL_BUT_SELF, MSG_RESCHEDULE, 0L, 0); */
#ifdef CONFIG_FAST_TIMER
	/* reset counter */
	miniinterrupts = FAST_FACTOR-1;
#endif
	    
	    
}

#ifndef	CONFIG_APM	/* cycle counter may be unreliable */

/* [From the Linux kernel development mailing list]
 *
 * 2.0 (i386) patch: "completely mad time"
 * 
 * Ulrich Windl (ulrich.windl@rz.uni-regensburg.de)
 * Tue, 16 Sep 1997 08:47:18 +0200
 * 
 * (Out of despair and lack of other good ideas I invented the "Complete Mad
 * Time (TM)":)
 * 
 * There's a difference with the timeoffset taken from the timer chip and
 * the fasttimeoffset taken from the Pentium cycle counter:
 * 
 * If a time interrupt is delayed with the Pentium stuff, the error
 * accumulates (e.g. when the first is 500us late and the next is
 * 200us late, the total delay is 700us). Due to the way the
 * fasttimeoffset is reset in the timer interrupt routine, these delays
 * are lost forever. I call this badness "1"
 * 
 * If a timer interrupt is delayed significantly (for more than one
 * tick), the current timeoffset is clamped at one tick. Thus the time
 * would no longer increase until the interrupt is processed. I call that
 * badness "2".
 * 
 * If you remove the clamping, the timeoffset can grow beyond one tick,
 * but on the next timer interrupt only one tick will be added and the
 * offset will be reset to zero, efectively makeing the time run
 * backwards. I call this badness "3".
 * 
 * I tried to fix these problems with my CMT (Completely Mad Time) patch
 * on top of 2.0.31-pre9 (The kernel should be tested anyway): The
 * clamping in the fast timeoffset has been removed, and the "cycles per
 * microsecond" is available more globally.
 * 
 * In the timer interrupt the delay is computed from the cycle counter
 * and added to the system clock. If the amount is larger than one tick
 * (very bad), only one tick is corrected and the remainder is
 * accumulated for later correction. If the delay is smaller, any
 * remaining correction from previous runs is subtracted from the system
 * time, but only up to one tick (because the time between the last to
 * invokations of the time routine was shorter than one tick then).
 * 
 * I haven't dealt with the APM case yet, and maybe some constants need
 * tuning (border cases). After all it looks not much worse than the
 * previous code. Is anyone out there that wants to test the patch? Ingo
 * Molnar still around?
 * 
 * There are two patches on top of 2.0.31-pre9 and against
 * arch/i386/kernel/time.c. Finally I have included a histogram of
 * measured time differences in a tight loop. Time doesn't look too
 * bad. The Program was taken from the xntp distribution (util/hist.c).
 * 
 * Have a nice time,
 * have a mad time,
 * have CMT
 * 
 * Ulrich
 */
/* @@ Note (Mark Carson):  A slightly subtle point with this code is that
 * no correction can actually made until the "cached_quotient" value is set,
 * which in turn only happens when do_gettimeofday is called.  But this will
 * happen with some frequency (e.g., called from ICMP code, and certainly
 * from NIST Net code!), so it shouldn't be a problem.  Of course, if no one
 * ever calls gettimeofday, it hardly matters what the clock reads....
 */

static void do_clock_correction(void)
{
       static  long    last_delay      = 0;
       register unsigned long eax asm("ax");
       register unsigned long edx asm("dx");
       unsigned long quotient = cached_quotient;
       unsigned long usecs;

       /* Read the time counter.. */
       __asm__(".byte 0x0f,0x31"
               :"=a" (eax), "=d" (edx));

       /* ..relative to previous interrupt */
       __asm__("subl "SYMBOL_NAME_STR(last_timer_cc)",%0\n\t"
               "sbbl "SYMBOL_NAME_STR(last_timer_cc)"+4,%1"
               :"=a" (eax), "=d" (edx)
               :"0" (eax), "1" (edx));

       /*
        * Check if we are late
        */
       __asm__("mull %2"
               :"=a" (eax), "=d" (edx)
               :"r" (quotient),
                "0" (eax), "1" (edx));
       usecs = edx;
       if (usecs > tick) {
               /* add the amount that we are late */
               if ((xtime.tv_usec += usecs - tick) > 1000000)
                       ++xtime.tv_sec, xtime.tv_usec -= 1000000;
               last_delay += usecs - tick;
       }
       else if ( last_delay > 0) {
               /* we are on schedule, but were late previously */
               /* as we should reduce tick now, we just reduce the time to
                  make it correct again */
               if ( last_delay > tick ) {
                       /* we have a slight problem here */
                       xtime.tv_usec -= tick;
                       last_delay -= tick;
               }
               else {
                       xtime.tv_usec -= last_delay;
                       last_delay = 0;
               }
               if (xtime.tv_usec < 0)
                       --xtime.tv_sec, xtime.tv_usec += 1000000;
               /* NOTE: Time won't go backwards (we'll add another tick) */

       }
}

/*
 * This is the same as the above, except we _also_ save the current
 * cycle counter value at the time of the timer interrupt, so that
 * we later on can estimate the time of day more exactly.
 */
static void pentium_timer_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
#ifdef CONFIG_FAST_TIMER
	/* Only record the cycle counter value once each jiffy */
	if (!miniinterrupts) {
#endif
	do_clock_correction();
	/* read Pentium cycle counter and let's forget that we were late
	 * (that's stored in last_delay)
	 */
	__asm__(".byte 0x0f,0x31"
		:"=a" (last_timer_cc.low),
		 "=d" (last_timer_cc.high));
#ifdef CONFIG_FAST_TIMER
	++local_lost_ticks;
	}
#endif
	timer_interrupt(irq, NULL, regs);
}
#endif

/* Converts Gregorian date to seconds since 1970-01-01 00:00:00.
 * Assumes input in normal date format, i.e. 1980-12-31 23:59:59
 * => year=1980, mon=12, day=31, hour=23, min=59, sec=59.
 *
 * [For the Julian calendar (which was used in Russia before 1917,
 * Britain & colonies before 1752, anywhere else before 1582,
 * and is still in use by some communities) leave out the
 * -year/100+year/400 terms, and add 10.]
 *
 * This algorithm was first published by Gauss (I think).
 *
 * WARNING: this function will overflow on 2106-02-07 06:28:16 on
 * machines were long is 32-bit! (However, as time_t is signed, we
 * will already get problems at other places on 2038-01-19 03:14:08)
 */
static inline unsigned long mktime(unsigned int year, unsigned int mon,
	unsigned int day, unsigned int hour,
	unsigned int min, unsigned int sec)
{
	if (0 >= (int) (mon -= 2)) {	/* 1..12 -> 11,12,1..10 */
		mon += 12;	/* Puts Feb last since it has leap day */
		year -= 1;
	}
	return (((
	    (unsigned long)(year/4 - year/100 + year/400 + 367*mon/12 + day) +
	      year*365 - 719499
	    )*24 + hour /* now have hours */
	   )*60 + min /* now have minutes */
	  )*60 + sec; /* finally seconds */
}

unsigned long get_cmos_time(void)
{
	unsigned int year, mon, day, hour, min, sec;
	int i;

	/* The Linux interpretation of the CMOS clock register contents:
	 * When the Update-In-Progress (UIP) flag goes from 1 to 0, the
	 * RTC registers show the second which has precisely just started.
	 * Let's hope other operating systems interpret the RTC the same way.
	 */
	/* read RTC exactly on falling edge of update flag */
	for (i = 0 ; i < 1000000 ; i++)	/* may take up to 1 second... */
		if (CMOS_READ(RTC_FREQ_SELECT) & RTC_UIP)
			break;
	for (i = 0 ; i < 1000000 ; i++)	/* must try at least 2.228 ms */
		if (!(CMOS_READ(RTC_FREQ_SELECT) & RTC_UIP))
			break;
	do { /* Isn't this overkill ? UIP above should guarantee consistency */
		sec = CMOS_READ(RTC_SECONDS);
		min = CMOS_READ(RTC_MINUTES);
		hour = CMOS_READ(RTC_HOURS);
		day = CMOS_READ(RTC_DAY_OF_MONTH);
		mon = CMOS_READ(RTC_MONTH);
		year = CMOS_READ(RTC_YEAR);
	} while (sec != CMOS_READ(RTC_SECONDS));
	if (!(CMOS_READ(RTC_CONTROL) & RTC_DM_BINARY) || RTC_ALWAYS_BCD)
	  {
	    BCD_TO_BIN(sec);
	    BCD_TO_BIN(min);
	    BCD_TO_BIN(hour);
	    BCD_TO_BIN(day);
	    BCD_TO_BIN(mon);
	    BCD_TO_BIN(year);
	  }
	if ((year += 1900) < 1970)
		year += 100;
	return mktime(year, mon, day, hour, min, sec);
}

static struct irqaction irq0  = { timer_interrupt, 0, 0, "timer", NULL, NULL};

void time_init(void)
{
	xtime.tv_sec = get_cmos_time();
	xtime.tv_usec = 0;

	/* If we have the CPU hardware time counters, use them */
#ifndef CONFIG_APM
				/* Don't use them if a suspend/resume could
                                   corrupt the timer value.  This problem
                                   needs more debugging. */
	if (x86_capability & 16) {
		do_gettimeoffset = do_fast_gettimeoffset;

		if( strcmp( x86_vendor_id, "AuthenticAMD" ) == 0 ) {
			if( x86 == 5 ) {
				if( x86_model == 0 ) {
					/* turn on cycle counters during power down */
					__asm__ __volatile__ (" movl $0x83, %%ecx \n \
								.byte 0x0f,0x32 \n \
								orl $1,%%eax \n \
								.byte 0x0f,0x30 \n " 
                                                                : : : "ax", "cx", "dx" );
					udelay(500);
				}
			}
		}

		/* read Pentium cycle counter */
		__asm__(".byte 0x0f,0x31"
			:"=a" (init_timer_cc.low),
			 "=d" (init_timer_cc.high));
		irq0.handler = pentium_timer_interrupt;
	}
#endif
	setup_x86_irq(0, &irq0);
}
